module Dry
  module Facts
    class Aggregate
      # object/hash, representing a state,
      # computed from many events
      # state computation rules

      @event_handlers = Hash.new

      attr_reader :events
      attr_accessor :id
      attr_accessor :uuid

      class << self
        def aggregate_data_from_event_types *event_klasses
          event_klasses.each do |klass|
            define_event_type_handler klass, :_transfer_data_from_event
          end
        end

        def build_all_from_events events
          events
          .group_by {|e|            e.aggregate_id }
          .map      {|_, g_events|  build_one_from_events(g_events) }
        end

        def build_one_from_events events
          self
          .new(events)
        end

        def define_event_type_handler klass, method_name
          @event_handlers ||= {}
          @event_handlers[klass.name] = method_name
        end

        def event_handlers
          @event_handlers
        end

        def event_types
          @event_handlers.keys
        end

        def find_event_handler(key)
          (@event_handlers)[key]
        end
      end

      def initialize(events)
        @events = events || []
        first_event_aggregate_id= if @events.first
          @events.first.aggregate_id
        end
        @uuid = @id = first_event_aggregate_id

        @events.each {|e| self.send :_handle_event, e}
      end

      def to_h
        self
        .instance_variables
        .map              {|ivar_name|  ivar_name.to_s.gsub(/@/, '')}
        .yield_self       {|ivars|      ivars - ['events', 'uuid']}
        .reduce(Hash.new) {|memo, obj|  memo[obj.to_sym] = self.send(obj); memo }
      end
      alias_method :to_hash, :to_h

      private

      def _default_aggregate_metadata
        uuid = SecureRandom.uuid,
        { id:   uuid,
          uuid: uuid,
          type: { name: self.class.name } }
      end

      def _handle_event event
        if self.uuid != event.metadata[:aggregate_id]
          fail "Event don't belong to aggregate instance. \n
                Aggregate #{self.inspect} \n
                Event: #{event.inspect}"
        end

        handler_name =
          self
          .class
          .find_event_handler(event.metadata[:type][:name])

        self.send(handler_name, event) if handler_name
      end

      def _transfer_data_from_event event
        event.data.keys.each do |key|
          if :id == key.to_sym
            self.send "#{key}=", event.aggregate_id
          else
            self.send "#{key}=", event.data[key]
          end
        end
      end

    end
  end
end
